"""OctoAI Image Generation."""
from __future__ import (
    annotations,  # required to allow 3.7+ python use type | syntax introduced in 3.10
)

from enum import Enum
from typing import Dict, List

from clients.image_gen.models import ImageEncoding, Scheduler, SDXLStyles
from octoai.clients.asset_orch import Asset
from octoai.clients.base_model import BaseModel
from octoai.errors import OctoAIValidationError
from octoai.types import Image

SDXL_ALLOWABLE_WIDTH_TO_HEIGHT = {
    1024: {1024, 768},
    896: {1152, 896},
    832: {1216, 512},
    768: {1344, 1024, 576, 512},
    704: {1216, 384},
    640: {1536, 768, 640},
    576: {1024, 448, 768},
    512: {832, 768, 704, 512},
    448: {576},
    384: {704},
    1536: {640},
    1344: {768},
    1216: {832, 704},
    1152: {896},
}

SD_ALLOWABLE_WIDTH_TO_HEIGHT = {
    576: {1024, 768},
    512: {512, 704, 768},
    640: {512, 768},
    768: {512},
    1024: {576},
}

SSD_ALLOWABLE_WIDTH_TO_HEIGHT = {
    640: {1536},
    768: {1344},
    832: {1216},
    896: {1152},
    1024: {1024},
    1152: {896},
    1216: {832},
    1344: {768},
    1536: {640},
}

MAX_CFG_SCALE = 50
MAX_NUM_IMAGES = 16
MAX_STEPS = 200


class Engine(str, Enum):
    """
    SDXL: Stable Diffusion XL
    SD: Stable Diffusion
    """

    SDXL = "sdxl"
    SD = "sd"
    SSD = "ssd"
    CONTROLNET_SDXL = "controlnet-sdxl"
    CONTROLNET_SD = "controlnet-sd"

    def to_name(self):
        if self == self.SDXL:
            return "Stable Diffusion XL"
        elif self == self.SD:
            return "Stable Diffusion"
        elif self == self.SSD:
            return "Stable Diffusion SSD"
        elif self == self.CONTROLNET_SDXL:
            return "ControlNet Stable Diffusion XL"
        elif self == self.CONTROLNET_SD:
            return "ControlNet Stable Diffusion 1.5"


class ImageGenerateResponse:
    """
    Image generation response.

    Contains a list of images as well as a counter of those filtered for safety.
    """

    def __init__(self, images: List[Image], removed_for_safety: int):
        self._images = images
        self._removed_for_safety = removed_for_safety

    @property
    def images(self) -> List[Image]:
        """Return list of :class:`Image` generated from request."""
        return self._images

    @property
    def removed_for_safety(self) -> int:
        """Return int representing number of images removed for safety."""
        return self._removed_for_safety


class ImageGenerator(BaseModel):
    """Client for image generation."""

    def __init__(
        self,
        api_endpoint: str | None = None,
        *args,
        **kwargs,
    ):
        if not api_endpoint:
            api_endpoint = "https://image.octoai.run/"

        super(ImageGenerator, self).__init__(api_endpoint, *args, **kwargs)

        if self._httpx_client.headers.get("Authorization") is None:
            msg = (
                "Authorization is required.  Please set an `OCTOAI_TOKEN` "
                "environment variable, or pass your token to the client using "
                "`client = ImageGenerator(token='your-octoai-api-token')`"
            )
            raise OctoAIValidationError(msg)

    # Raises OctoAIValidationError on failure to validate variable.
    # Does not validate strings currently, though should once API is stable.
    # TODO: standardize error strings (once more input types are known).
    def _validate_inputs(
        self,
        engine: Engine | str,
        cfg_scale: float | None,
        height: int | None,
        high_noise_frac: float | None,
        num_images: int | None,
        seed: int | None,
        steps: int | None,
        strength: float | None,
        width: int | None,
        image_encoding: ImageEncoding | str | None,
        sampler: Scheduler | str | None,
        prompt_2: str | None,
        negative_prompt_2: str | None,
        use_refiner: bool | None,
        init_image: str | None = None,
        controlnet: str | None = None,
        controlnet_image: str | None = None,
        controlnet_conditioning_scale: float | None = None,
        style_preset: str | None = None,
    ):
        """Validate inputs."""
        engines = [e.value for e in Engine]
        if engine not in engines:
            raise OctoAIValidationError(
                f"engine set to {engine}.  Must be one of: {', '.join(engines)}."
            )

        # Check only compatible with engine attributes being used
        if high_noise_frac and engine == "sd":
            self._input_not_match_engine(
                "high_noise_frac", high_noise_frac, engine, "sdxl"
            )
        if prompt_2 and engine == "sd":
            self._input_not_match_engine("prompt_2", prompt_2, engine, "sdxl")
        if negative_prompt_2 and engine == "sd":
            self._input_not_match_engine(
                "negative_prompt_2", negative_prompt_2, engine, "sdxl"
            )
        if use_refiner is not None and engine == "sd":
            self._input_not_match_engine("use_refiner", use_refiner, engine, "sdxl")

        # Check number values in range
        if cfg_scale is not None and ((0 > cfg_scale) or (MAX_CFG_SCALE < cfg_scale)):
            raise OctoAIValidationError(
                f"cfg_scale set to: {cfg_scale}. Allowable range is > 0 and <= {MAX_CFG_SCALE}."
            )

        if high_noise_frac is not None and (0 > high_noise_frac or high_noise_frac > 1):
            raise OctoAIValidationError(
                f"high_noise_frac set to: {high_noise_frac}. Allowable range is "
                f">= 0 and <= 1."
            )
        if num_images is not None and (0 >= num_images or num_images > MAX_NUM_IMAGES):
            raise OctoAIValidationError(
                f"num_images set to: {num_images}.  Allowable range is > 0 and <= {MAX_NUM_IMAGES}."
            )
        if isinstance(seed, list):
            for each in seed:
                if each is not None and (0 > each or each >= 2**32):
                    raise OctoAIValidationError(
                        f"seed({seed}) contains {each}. Allowable range is >= 0 and "
                        f"< 2**32."
                    )
        if type(seed) == int and (0 > seed or seed >= 2**32):
            raise OctoAIValidationError(
                f"seed set to: {seed}.  Allowable range is >= 0 and < 2**32."
            )
        if steps is not None and (0 >= steps or steps > MAX_STEPS):
            raise OctoAIValidationError(
                f"steps set to: {steps}.  Allowable range is > 0 and <= {MAX_STEPS}."
            )
        if strength is not None and (0 > strength or strength > 1):
            raise OctoAIValidationError(
                f"strength set to: {strength}.  Allowable range is >= 0 and <= 1."
            )
        if (
            controlnet_conditioning_scale is not None
            and controlnet_conditioning_scale < 0
        ):
            raise OctoAIValidationError(
                f"controlnet_conditional_scale set to: {controlnet_conditioning_scale}."
                " Allowable range is >= 0."
            )

        # More verifying required value
        if strength is not None and init_image is None:
            raise OctoAIValidationError(
                f"init_image required for img2img generation.  "
                f"strength({strength}) cannot be set if "
                f"init_image is None."
            )

        if style_preset is not None and engine == "sd":
            self._input_not_match_engine("style_preset", style_preset, engine, "sdxl")
        if style_preset is not None and engine == "sdxl":
            try:
                style_preset = SDXLStyles(style_preset)
            except ValueError:
                msg = (
                    f"style_preset({style_preset}) is not valid.  "
                    f"Valid options include {[e.value for e in SDXLStyles]}."
                )
                raise OctoAIValidationError(msg)

        # Validate width and height to engine
        if (height is None) ^ (width is None):
            raise OctoAIValidationError(
                f"if height({height}) or width({width}) is set "
                f"to None, both must be None."
            )
        engine = Engine(engine)
        if height is not None and width is not None:
            self._validate_height_and_width_to_engine(engine, height, width)

        if image_encoding is not None:
            try:
                image_encoding = ImageEncoding(image_encoding)
            except ValueError:
                msg = (
                    f"image_encoding set to {image_encoding}.  "
                    f"Valid options include {[e.value for e in ImageEncoding]}."
                )
                raise OctoAIValidationError(msg)

        if sampler is not None:
            try:
                sampler = Scheduler(sampler)
            except ValueError:
                msg = (
                    f"sampler set to {sampler}.  Valid options include "
                    f"{[e.value for e in Scheduler]}."
                )
                raise OctoAIValidationError(msg)

        if init_image is not None:
            if not isinstance(init_image, Image):
                init_image = Image(init_image)
            if not init_image.is_valid():
                msg = (
                    "init_image is not a valid image.  May either use the "
                    "octoai.types Image class or a base64 string."
                )
                raise OctoAIValidationError(msg)

        if controlnet_image is not None:
            if not isinstance(controlnet_image, Image):
                controlnet_image = Image(controlnet_image)
            if not controlnet_image.is_valid():
                msg = (
                    "controlnet_image is not a valid image.  May either use the "
                    "octoai.types Image class or a base64 string."
                )
                raise OctoAIValidationError(msg)

        if engine in [
            Engine.CONTROLNET_SDXL,
            Engine.CONTROLNET_SDXL.value,
            Engine.CONTROLNET_SD,
            Engine.CONTROLNET_SD.value,
        ]:
            if controlnet_image is None:
                raise OctoAIValidationError(
                    f"controlnet_image is required for engine {engine}."
                )

            if controlnet is None:
                raise OctoAIValidationError(
                    f"controlnet is required for engine {engine}."
                )

    # Server will return a 500 error if incorrect height and width are entered
    @staticmethod
    def _validate_height_and_width_to_engine(engine: Engine, height: int, width: int):
        width_to_height_by_engine = {
            Engine.SDXL.value: SDXL_ALLOWABLE_WIDTH_TO_HEIGHT,
            Engine.SD.value: SD_ALLOWABLE_WIDTH_TO_HEIGHT,
            Engine.SSD.value: SSD_ALLOWABLE_WIDTH_TO_HEIGHT,
            Engine.CONTROLNET_SDXL.value: SDXL_ALLOWABLE_WIDTH_TO_HEIGHT,
            Engine.CONTROLNET_SD.value: SD_ALLOWABLE_WIDTH_TO_HEIGHT,
        }

        # Set to correct const allowable values
        width_to_height = width_to_height_by_engine.get(engine)
        if width_to_height.get(width) is None:
            raise OctoAIValidationError(
                f"width ({width}): height ({height}) "
                f"values must match {engine.to_name()} allowable values or both be "
                f"None. Valid values for width are {list(width_to_height.keys())}."
            )
        if height not in width_to_height.get(width):
            raise OctoAIValidationError(
                f"width ({width}): height ({height}) "
                f"values must match {engine.to_name()} allowable values or both be "
                f"None. Valid width:height values are: {width_to_height}."
            )

    @staticmethod
    def _input_not_match_engine(name, value, engine, ok_engine):
        raise OctoAIValidationError(
            f"{name}({value}) is set but engine is set to "
            f"{engine}. {name} can only be used with {ok_engine}."
        )

    # May be worthwhile to limit steps to 50, but at this time, 100 matches the UI
    # num_images limit of 10 temp set by Itay, but part of larger limit
    # discussion.
    def generate(
        self,
        engine: Engine | str,
        prompt: str,
        prompt_2: str | None = None,  # SDXL only
        negative_prompt: str | None = None,
        negative_prompt_2: str | None = None,  # SDXL only
        checkpoint: str | Asset | None = None,
        vae: str | Asset | None = None,
        textual_inversions: Dict[str | Asset, str] | None = None,
        loras: Dict[str | Asset, float] | None = None,
        sampler: str | Scheduler | None = None,  # Server default DDIM
        height: int | None = None,  # Different defaults for sdxl and sd
        width: int | None = None,
        cfg_scale: float | None = 12.0,
        steps: int | None = 30,
        num_images: int | None = 1,
        seed: int | List[int] | None = None,
        init_image: str | Image | None = None,  # b64, img2img only
        controlnet: str | None = None,  # controlnet-sdxl|sd only
        controlnet_image: str | Image | None = None,  # b64, controlnet-sdxl|sd only
        controlnet_conditioning_scale: float
        | None = None,  # 1.0 server default, controlnet-sdxl|sd only
        strength: float | None = None,  # 0.8 server default, img2img only
        style_preset: str | SDXLStyles | None = None,
        use_refiner: bool | None = None,  # True default, SDXL only
        high_noise_frac: float | None = None,  # 0.8 server default, SDXL only
        enable_safety: bool | None = True,
        image_encoding: ImageEncoding | str | None = ImageEncoding.JPEG,
    ) -> ImageGenerateResponse:
        """
        Generate a list of images based on request.

        :param engine: Required. "sdxl" for Stable Diffusion XL; "sd" for Stable
            Diffusion 1.5; "ssd" for Stable Diffusion SSD; "controlnet-sdxl" for
            ControlNet Stable Diffusion XL; "controlnet-sd" for ControlNet Stable
            Diffusion 1.5.
        :param prompt: Required. Describes the image to generate.
            ex. "An octopus playing chess, masterpiece, photorealistic"
        :param prompt_2: High level description of the image to generate,
            defaults to None.
        :param negative_prompt: Description of image traits to avoid, defaults to None.
            ex. "Fingers, distortions"
        :param negative_prompt_2: High level description of things to avoid during
            generation, defaults to None.
            ex. "Unusual proportions and distorted faces"
        :param checkpoint: Which checkpoint to use for inferences, defaults to None.
        :param vae: Custom VAE to be used during image generation, defaults to None.
        :param textual_inversions: A dictionary of textual inversion updates,
            defaults to None
            ex. {'name': 'trigger_word'}
        :param loras: A dictionary of LoRAs updates to apply and their weight, can also
            be used with Assets created in the SDK directly, defaults to None.
            ex. {'crayon-style': 0.3, my_created_asset: 0.1}
        :param sampler: :class:`Scheduler` to use when generating image,
            defaults to None.
        :param height: Height of image to generate, defaults to None.
        :param width: Width of image to generate, defaults to None.
        :param cfg_scale: How closely to adhere to prompt description, defaults to 12.0.
            Must be >= 0 and <= 50.
        :param steps: How many steps of diffusion to run, defaults to 30.
            May be > 0 and <= 100.
        :param num_images: How many images to generate, defaults to 1.
            May be > 0 and <= 4.
        :param seed: Fixed random seed, useful when attempting to generate a
            specific image, defaults to None.
            May be >= 0 < 2**32.
        :param init_image: Starting image for img2img mode, defaults to None.
            Requires a b64 string image or :class:`Image`.
        :param controlnet: String matching id of controlnet to use for controlnet engine
            inferences, defaults to None.  Required for using controlnet engines.
        :param controlnet_image: Starting image for controlnet-sdxl mode, defaults to None.
            Requires a b64 string image or :class:`Image`.
        :param controlnet_conditioning_scale: How strong the effect of the controlnet
            should be, defaults to 1.0.
        :param strength: How much creative to be in img2img mode, defaults to 0.8.
            May be >= 0 and <= 1.  Must have an `init_image`.
        :param style_preset: Used to guide the output image towards a particular style,
            only usable with SDXL,defaults to None.
            ex. "low-poly"
        :param use_refiner: Whether to apply the sdxl refiner, defaults to True.
        :param high_noise_frac: Which fraction of steps to perform with the base model,
            defaults to 0.8.
            May be >= 0 and <= 1.
        :param enable_safety: Whether to use safety checking on generated outputs or
            not, defaults to True.
        :param image_encoding: Choose returned :class:`ImageEncoding` type,
            defaults to :class:`ImageEncoding.JPEG`.
        :return: :class:`GenerateImagesResponse` object including properties for a list
            of images as well as a counter of total images returned below the
            `num_images` value due to being removed for safety.
        """
        self._validate_inputs(
            engine,
            cfg_scale,
            height,
            high_noise_frac,
            num_images,
            seed,
            steps,
            strength,
            width,
            image_encoding,
            sampler,
            prompt_2,
            negative_prompt_2,
            use_refiner,
            init_image,
            controlnet,
            controlnet_image,
            controlnet_conditioning_scale,
            style_preset,
        )
        if isinstance(init_image, Image):
            init_image = init_image.to_base64()
        if isinstance(controlnet_image, Image):
            controlnet_image = controlnet_image.to_base64()

        inputs = self._process_local_vars_to_inputs_dict(locals())

        images = []
        endpoint = self.api_endpoint + "generate/" + engine
        output = self.infer(endpoint, inputs)
        removed_for_safety = 0
        for image_b64 in output.get("images"):
            if image_b64.get("removed_for_safety"):
                removed_for_safety += 1
            else:
                image_b64_str = image_b64.get("image_b64")
                image = Image(image_b64_str)
                images.append(image)
        return ImageGenerateResponse(
            images=images, removed_for_safety=removed_for_safety
        )

    # If the key is of type str, it does nothing, otherwise it returns a dict where
    # objects with id fields have the id field used as the key instead.
    # Examples are Assets as keys.
    @staticmethod
    def _replace_object_keys_with_ids(obj_dict: dict) -> dict:
        result = {}
        for key, value in obj_dict.items():
            if isinstance(key, Asset):
                key = key.id
            elif type(key) != str:
                msg = (
                    f"key({key}) is invalid.  `loras` and `textual_inversions` "
                    f"require keys in dictionary to either be a str "
                    f"or :class:`Asset`."
                )
                raise OctoAIValidationError(msg)
            result[key] = value
        return result

    # Purges irrelevant locals from inputs dict and converts Asset type to ids
    def _process_local_vars_to_inputs_dict(self, inputs):
        inputs.pop("self")
        for key in list(inputs):
            if inputs[key] is None:
                inputs.pop(key)
        inputs.pop("engine")
        if "loras" in inputs:
            inputs["loras"] = self._replace_object_keys_with_ids(inputs["loras"])
        if "textual_inversions" in inputs:
            inputs["textual_inversions"] = self._replace_object_keys_with_ids(
                inputs["textual_inversions"]
            )
        if isinstance(inputs.get("vae"), Asset):
            inputs["vae"] = inputs["vae"].id
        if isinstance(inputs.get("checkpoint"), Asset):
            inputs["checkpoint"] = inputs["checkpoint"].id
        return inputs



# coding: utf-8

# flake8: noqa
"""
    fine-tuning-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from clients.fine_tuning.models.base_engine import BaseEngine
from clients.fine_tuning.models.create_tune_request import CreateTuneRequest
from clients.fine_tuning.models.details import Details
from clients.fine_tuning.models.details1 import Details1
from clients.fine_tuning.models.http_validation_error import HTTPValidationError
from clients.fine_tuning.models.list_tunes_response import ListTunesResponse
from clients.fine_tuning.models.location_inner import LocationInner
from clients.fine_tuning.models.lora_tune_checkpoint import LoraTuneCheckpoint
from clients.fine_tuning.models.lora_tune_file import LoraTuneFile
from clients.fine_tuning.models.lora_tune_input import LoraTuneInput
from clients.fine_tuning.models.lora_tune_output import LoraTuneOutput
from clients.fine_tuning.models.null_tune import NullTune
from clients.fine_tuning.models.tune import Tune
from clients.fine_tuning.models.tune_status import TuneStatus
from clients.fine_tuning.models.validation_error import ValidationError



"""
Contains the base interface that OctoAI endpoints should implement.

Developers that want to create an endpoint should implement the
``Service`` class in this module as directed by the ``octoai`` command-line
interface.
"""
import functools
import inspect
import os
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Callable, Mapping, Tuple, Type

import pydantic_core
from fastapi import Form
from pydantic import BaseModel, Field, create_model

from .types import File

DEFAULT_VOLUME_PATH = "/octoai/cache"
STORE_ASSETS_NOT_OVERRIDDEN = "NOT_OVERRIDDEN"


def volume_path() -> str:
    """Get mounted volume path in docker.

    :return: Docker path.
    """
    docker_path = os.environ.get("OCTOAI_VOLUME_PATH", None)
    if docker_path:
        return docker_path
    else:
        return DEFAULT_VOLUME_PATH


VOLUME_ENVIRONMENT = {
    "HUGGINGFACE_HUB_CACHE": os.path.join(volume_path(), "huggingface_hub_cache"),
    "TORCH_HOME": os.path.join(volume_path(), "torch_home"),
}


class Service(ABC):
    """
    The base interface that OctoAI endpoints should implement.

    Developers that want to create an endpoint should implement this
    class as directed by the ``octoai`` command-line interface.
    """

    def setup(self) -> None:
        """
        Perform service initialization.

        A common operation to include here is loading weights and making
        those available to the ``infer()`` method in a member variable.
        """
        pass

    def store_assets(self) -> None:
        """Download model assets."""
        pass

    def on_server_startup(self) -> None:
        """
        Perform any necessary initialization when the server starts.

        This method is separate from setup() because setup() can be called
        outside the serving context to include weights when building the image.
        """
        pass

    def on_server_shutdown(self) -> None:
        """Perform any necessary cleanup when the server stops."""
        pass

    @abstractmethod
    def infer(self, **kwargs: Any) -> Any:
        """Perform inference."""
        pass

    setattr(store_assets, STORE_ASSETS_NOT_OVERRIDDEN, True)


class ResponseAnalytics(BaseModel):
    """Additional analytics metadata."""

    inference_time_ms: float = Field(
        description="Inference execution time (without pauses)"
    )
    performance_time_ms: float = Field(
        description="Inference execution time (including pauses)"
    )


def inspect_input_types(method: Callable) -> Type[BaseModel]:
    """Create Pydantic input model from inference method signature."""
    signature = inspect.signature(method)

    if len(signature.parameters) < 1:
        raise ValueError(f"{method.__name__}() requires at least one argument")

    return inspect_input_parameters(signature.parameters)


def inspect_input_parameters(
    parameters: Mapping[str, inspect.Parameter]
) -> Type[BaseModel]:
    """Create Pydnatic input mode from the provided signature."""
    args = OrderedDict()

    for arg_name, arg in parameters.items():
        if arg.annotation == inspect.Parameter.empty:
            raise ValueError("infer() requires type annotations for args")

        default = (
            pydantic_core.PydanticUndefined
            if arg.default == inspect.Parameter.empty
            else arg.default
        )
        args[arg_name] = (
            arg.annotation,
            Field(default=default),
        )

    return create_model(
        "Input",
        __config__=None,
        __base__=BaseModel,
        __module__=__name__,
        __validators__=None,
        **args,
    )


def inspect_output_types(method: Callable) -> Type[BaseModel]:
    """Create Pydantic output model from inference method signature."""
    signature = inspect.signature(method)

    if signature.return_annotation == inspect._empty:
        raise ValueError(f"{method.__name__}() requires a return type annotation")

    rets = OrderedDict()
    rets["output"] = (signature.return_annotation, None)
    rets["analytics"] = (ResponseAnalytics, None)

    return create_model(
        "Output",
        __config__=None,
        __base__=BaseModel,
        __module__=__name__,
        __validators__=None,
        **rets,
    )


def implements_form_data(service: Service) -> bool:
    """Check if ``infer_form_data()`` is implemented."""
    for name, method in inspect.getmembers(service, predicate=inspect.ismethod):
        if name == "infer_form_data":
            return True

    return False


def transform_form_data_signature(service: Service) -> inspect.Signature:
    """Add FastAPI Form() defaults to 'str' fields and check compatibility."""
    if not hasattr(service.infer_form_data, "__path__"):  # type: ignore[attr-defined]
        raise ValueError(
            "infer_form_data() must be decorated with @path() "
            "when using form-data inference."
        )

    if service.infer_form_data.__path__ == "/infer":  # type: ignore[attr-defined]
        raise ValueError("Cannot use /infer for form-data inference.")

    params = []
    signature = inspect.signature(service.infer_form_data)  # type: ignore[attr-defined]
    for param in signature.parameters.values():
        if param.annotation == File:
            params.append(param)
        elif param.annotation == str:
            params.append(param.replace(default=Form()))
        else:
            raise ValueError("Only File and str params are supported for form data.")

    if not params:
        raise ValueError("infer_form_data() must have at least one parameter.")

    return inspect.Signature(params)


def find_additional_endpoints(service: Service) -> Mapping[str, Tuple[Callable, str]]:
    """Find additional methods that should be exposed as endpoints."""
    methods = {}
    reserved_methods = [attr for attr in dir(Service) if not attr.startswith("_")]
    reserved_methods.append("infer_form_data")

    for name, method in inspect.getmembers(service, predicate=inspect.ismethod):
        if name not in reserved_methods and not name.startswith("_"):
            methods[name] = (method, getattr(method, "__path__", None))

    return methods


def path(path: str):
    """Specify the path for a service method."""

    def wrapped(fn: Callable):
        @functools.wraps(fn)
        def wrapped_fn(*args, **kwargs):
            return fn(*args, **kwargs)

        wrapped_fn.__path__ = path  # type: ignore[attr-defined]
        return wrapped_fn

    return wrapped



# coding: utf-8

"""
    FastAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic.v1 import BaseModel, Field, StrictBool, StrictInt, StrictStr

class ImageGeneration(BaseModel):
    """
    Represent a generated image.  # noqa: E501
    """
    image_b64: Optional[StrictStr] = Field(...)
    removed_for_safety: StrictBool = Field(..., description="True if the image was removed for safety, False otherwise.")
    seed: StrictInt = Field(..., description="Seed used to generate the image. It can be used to recreate the generation.")
    __properties = ["image_b64", "removed_for_safety", "seed"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ImageGeneration:
        """Create an instance of ImageGeneration from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # set to None if image_b64 (nullable) is None
        # and __fields_set__ contains the field
        if self.image_b64 is None and "image_b64" in self.__fields_set__:
            _dict['image_b64'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ImageGeneration:
        """Create an instance of ImageGeneration from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ImageGeneration.parse_obj(obj)

        _obj = ImageGeneration.parse_obj({
            "image_b64": obj.get("image_b64"),
            "removed_for_safety": obj.get("removed_for_safety"),
            "seed": obj.get("seed")
        })
        return _obj





# coding: utf-8

"""
    FastAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, List, Optional, Union
from pydantic.v1 import BaseModel, Field, StrictBool, StrictFloat, StrictInt, StrictStr, conlist

class LegacyImageGenerationResponse(BaseModel):
    """
    Legacy model response.  # noqa: E501
    """
    completion: Optional[Any] = Field(...)
    prediction_time_ms: Union[StrictFloat, StrictInt] = Field(...)
    replica: StrictStr = Field(...)
    seed: StrictInt = Field(...)
    removed_for_safety: conlist(StrictBool) = Field(...)
    __properties = ["completion", "prediction_time_ms", "replica", "seed", "removed_for_safety"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> LegacyImageGenerationResponse:
        """Create an instance of LegacyImageGenerationResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # set to None if completion (nullable) is None
        # and __fields_set__ contains the field
        if self.completion is None and "completion" in self.__fields_set__:
            _dict['completion'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> LegacyImageGenerationResponse:
        """Create an instance of LegacyImageGenerationResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return LegacyImageGenerationResponse.parse_obj(obj)

        _obj = LegacyImageGenerationResponse.parse_obj({
            "prediction_time_ms": obj.get("prediction_time_ms"),
            "replica": obj.get("replica"),
            "seed": obj.get("seed"),
            "removed_for_safety": obj.get("removed_for_safety")
        })
        return _obj





"""Initializes the octoai module."""
from importlib.metadata import version

from . import chat, clients, errors, types, utils

__version__ = version("octoai-sdk")
__all__ = [
    "chat",
    "clients",
    "client",
    "errors",
    "server",
    "service",
    "types",
    "utils",
]



# coding: utf-8

"""
    FastAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Dict, List, Optional, Union
from pydantic.v1 import BaseModel, Field, StrictBool, StrictFloat, StrictInt, StrictStr, conlist
from clients.ollm.models.chat_completion_request_ext import ChatCompletionRequestExt
from clients.ollm.models.chat_message import ChatMessage
from clients.ollm.models.function import Function
from clients.ollm.models.function_call import FunctionCall
from clients.ollm.models.stop import Stop

class CreateChatCompletionRequest(BaseModel):
    """
    An OpenAI API compatible schema for a chat completion request.  See: https://platform.openai.com/docs/api-reference/chat/create for more information.  # noqa: E501
    """
    model: StrictStr = Field(...)
    messages: conlist(ChatMessage) = Field(...)
    functions: Optional[conlist(Function)] = None
    function_call: Optional[FunctionCall] = None
    temperature: Optional[Union[StrictFloat, StrictInt]] = 1.0
    top_p: Optional[Union[StrictFloat, StrictInt]] = 1.0
    n: Optional[StrictInt] = 1
    stream: Optional[StrictBool] = False
    stop: Optional[Stop] = None
    max_tokens: Optional[Union[StrictFloat, StrictInt]] = 262144
    presence_penalty: Optional[Union[StrictFloat, StrictInt]] = 0.0
    frequency_penalty: Optional[Union[StrictFloat, StrictInt]] = 0.0
    logit_bias: Optional[Dict[str, StrictStr]] = None
    user: Optional[StrictStr] = None
    octoai: Optional[ChatCompletionRequestExt] = None
    __properties = ["model", "messages", "functions", "function_call", "temperature", "top_p", "n", "stream", "stop", "max_tokens", "presence_penalty", "frequency_penalty", "logit_bias", "user", "octoai"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> CreateChatCompletionRequest:
        """Create an instance of CreateChatCompletionRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of each item in messages (list)
        _items = []
        if self.messages:
            for _item in self.messages:
                if _item:
                    _items.append(_item.to_dict())
            _dict['messages'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in functions (list)
        _items = []
        if self.functions:
            for _item in self.functions:
                if _item:
                    _items.append(_item.to_dict())
            _dict['functions'] = _items
        # override the default output from pydantic by calling `to_dict()` of function_call
        if self.function_call:
            _dict['function_call'] = self.function_call.to_dict()
        # override the default output from pydantic by calling `to_dict()` of stop
        if self.stop:
            _dict['stop'] = self.stop.to_dict()
        # override the default output from pydantic by calling `to_dict()` of octoai
        if self.octoai:
            _dict['octoai'] = self.octoai.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> CreateChatCompletionRequest:
        """Create an instance of CreateChatCompletionRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return CreateChatCompletionRequest.parse_obj(obj)

        _obj = CreateChatCompletionRequest.parse_obj({
            "model": obj.get("model"),
            "messages": [ChatMessage.from_dict(_item) for _item in obj.get("messages")] if obj.get("messages") is not None else None,
            "functions": [Function.from_dict(_item) for _item in obj.get("functions")] if obj.get("functions") is not None else None,
            "function_call": FunctionCall.from_dict(obj.get("function_call")) if obj.get("function_call") is not None else None,
            "temperature": obj.get("temperature") if obj.get("temperature") is not None else 1.0,
            "top_p": obj.get("top_p") if obj.get("top_p") is not None else 1.0,
            "n": obj.get("n") if obj.get("n") is not None else 1,
            "stream": obj.get("stream") if obj.get("stream") is not None else False,
            "stop": Stop.from_dict(obj.get("stop")) if obj.get("stop") is not None else None,
            "max_tokens": obj.get("max_tokens") if obj.get("max_tokens") is not None else 262144,
            "presence_penalty": obj.get("presence_penalty") if obj.get("presence_penalty") is not None else 0.0,
            "frequency_penalty": obj.get("frequency_penalty") if obj.get("frequency_penalty") is not None else 0.0,
            "logit_bias": obj.get("logit_bias"),
            "user": obj.get("user"),
            "octoai": ChatCompletionRequestExt.from_dict(obj.get("octoai")) if obj.get("octoai") is not None else None
        })
        return _obj





"""Example OctoAI service scaffold: Hello World."""
from octoai.service import Service


class HelloService(Service):
    """An OctoAI service extends octoai.service.Service."""

    # Uncomment this function if you want to place large assets
    # (checkpoints, models, etc) on a docker volume instead of building
    # them into the image. This can provide faster
    # cold-start time. You may need to modify your inference pipeline
    # so that it leverages the locally-stored assets instead of
    # downloading them at container boot time.
    # def store_assets(self) -> None:
    #     """Download model assets."""
    #     # By default, instantiate the model to fill local disk
    #     # caches on the OctoAI volume. You can customize this if
    #     # your model can be fetched without instantiating it.
    #     self.setup()

    def setup(self):
        """Perform intialization."""
        print("Setting up.")

    def infer(self, prompt: str) -> str:
        """Perform inference."""
        return prompt



# coding: utf-8

"""
    asset-lake-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, Optional
from pydantic.v1 import BaseModel, Field
from clients.asset_lake.models.archiver_volume_storage_ref import ArchiverVolumeStorageRef

class StorageDescriptor(BaseModel):
    """
    StorageDescriptor
    """
    archiver_storage_ref: ArchiverVolumeStorageRef = Field(...)
    storage_format: Optional[Any] = Field(...)
    __properties = ["archiver_storage_ref", "storage_format"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> StorageDescriptor:
        """Create an instance of StorageDescriptor from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of archiver_storage_ref
        if self.archiver_storage_ref:
            _dict['archiver_storage_ref'] = self.archiver_storage_ref.to_dict()
        # set to None if storage_format (nullable) is None
        # and __fields_set__ contains the field
        if self.storage_format is None and "storage_format" in self.__fields_set__:
            _dict['storage_format'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> StorageDescriptor:
        """Create an instance of StorageDescriptor from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return StorageDescriptor.parse_obj(obj)

        _obj = StorageDescriptor.parse_obj({
            "archiver_storage_ref": ArchiverVolumeStorageRef.from_dict(obj.get("archiver_storage_ref")) if obj.get("archiver_storage_ref") is not None else None,
            "storage_format": obj.get("storage_format")
        })
        return _obj





# coding: utf-8

"""
    FastAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, Optional
from pydantic.v1 import BaseModel, Field
from clients.image_gen.models.checkpoint import Checkpoint
from clients.image_gen.models.control_net import ControlNet
from clients.image_gen.models.control_net_image import ControlNetImage
from clients.image_gen.models.dictionary_of_lo_ras import DictionaryOfLoRAs
from clients.image_gen.models.dictionary_of_textual_inversions import DictionaryOfTextualInversions
from clients.image_gen.models.fixed_random_seed import FixedRandomSeed
from clients.image_gen.models.image_encoding import ImageEncoding
from clients.image_gen.models.initial_image import InitialImage
from clients.image_gen.models.mask_image import MaskImage
from clients.image_gen.models.negative_prompt import NegativePrompt
from clients.image_gen.models.output_image_height import OutputImageHeight
from clients.image_gen.models.output_image_width import OutputImageWidth
from clients.image_gen.models.pre_defined_styles import PreDefinedStyles
from clients.image_gen.models.scheduler import Scheduler
from clients.image_gen.models.second_input_prompt import SecondInputPrompt
from clients.image_gen.models.second_negative_prompt import SecondNegativePrompt
from clients.image_gen.models.vae import VAE

class ImageGenerationRequest(BaseModel):
    """
    Generate one or more images based on the given parameters.  # noqa: E501
    """
    prompt: Optional[Any] = Field(..., description="Text describing the image content to generate.")
    prompt_2: Optional[SecondInputPrompt] = None
    negative_prompt: Optional[NegativePrompt] = None
    negative_prompt_2: Optional[SecondNegativePrompt] = None
    checkpoint: Optional[Checkpoint] = None
    controlnet: Optional[ControlNet] = None
    vae: Optional[VAE] = None
    textual_inversions: Optional[DictionaryOfTextualInversions] = None
    loras: Optional[DictionaryOfLoRAs] = None
    sampler: Optional[Scheduler] = None
    height: Optional[OutputImageHeight] = None
    width: Optional[OutputImageWidth] = None
    cfg_scale: Optional[Any] = Field(None, description="Floating-point number represeting how closely to adhere to prompt description. Must be a positive number no greater than 50.0.")
    steps: Optional[Any] = Field(None, description="Integer repreenting how many steps of diffusion to run. Must be greater than 0 and less than or equal to 200.")
    num_images: Optional[Any] = Field(None, description="Integer representing how many output images to generate with a single prompt/configuration.")
    seed: Optional[FixedRandomSeed] = None
    controlnet_image: Optional[ControlNetImage] = None
    init_image: Optional[InitialImage] = None
    mask_image: Optional[MaskImage] = None
    strength: Optional[Any] = Field(None, description="Floating-point number indicating how much creative the Image to Image generation mode should be. Must be greater than 0 and less than or equal to 1.0.")
    style_preset: Optional[PreDefinedStyles] = None
    use_refiner: Optional[Any] = Field(None, description="Whether to enable and apply the SDXL refiner model to the image generation.")
    high_noise_frac: Optional[Any] = Field(None, description="Floating-point number that defines the fraction of steps to perform with the base model. Used only by SD XL. Must be greater than or equal to 0.0 and less than or equal to 1.0.")
    controlnet_conditioning_scale: Optional[Any] = Field(None, description="How strong the effect of the controlnet should be.")
    enable_safety: Optional[Any] = Field(None, description="Boolean defining whether to use safety checker system on generated outputs or not.")
    image_encoding: Optional[ImageEncoding] = None
    force_asset_download: Optional[Any] = Field(None, description="[Internal] Boolean defining if assets must be re-downloaded into the cache even if present.")
    force_asset_gpu_copy: Optional[Any] = Field(None, description="[Internal] Boolean defining if assets must to be copied into the GPU even if present.")
    additional_properties: Dict[str, Any] = {}
    __properties = ["prompt", "prompt_2", "negative_prompt", "negative_prompt_2", "checkpoint", "controlnet", "vae", "textual_inversions", "loras", "sampler", "height", "width", "cfg_scale", "steps", "num_images", "seed", "controlnet_image", "init_image", "mask_image", "strength", "style_preset", "use_refiner", "high_noise_frac", "controlnet_conditioning_scale", "enable_safety", "image_encoding", "force_asset_download", "force_asset_gpu_copy"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ImageGenerationRequest:
        """Create an instance of ImageGenerationRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                            "additional_properties"
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of prompt_2
        if self.prompt_2:
            _dict['prompt_2'] = self.prompt_2.to_dict()
        # override the default output from pydantic by calling `to_dict()` of negative_prompt
        if self.negative_prompt:
            _dict['negative_prompt'] = self.negative_prompt.to_dict()
        # override the default output from pydantic by calling `to_dict()` of negative_prompt_2
        if self.negative_prompt_2:
            _dict['negative_prompt_2'] = self.negative_prompt_2.to_dict()
        # override the default output from pydantic by calling `to_dict()` of checkpoint
        if self.checkpoint:
            _dict['checkpoint'] = self.checkpoint.to_dict()
        # override the default output from pydantic by calling `to_dict()` of controlnet
        if self.controlnet:
            _dict['controlnet'] = self.controlnet.to_dict()
        # override the default output from pydantic by calling `to_dict()` of vae
        if self.vae:
            _dict['vae'] = self.vae.to_dict()
        # override the default output from pydantic by calling `to_dict()` of textual_inversions
        if self.textual_inversions:
            _dict['textual_inversions'] = self.textual_inversions.to_dict()
        # override the default output from pydantic by calling `to_dict()` of loras
        if self.loras:
            _dict['loras'] = self.loras.to_dict()
        # override the default output from pydantic by calling `to_dict()` of height
        if self.height:
            _dict['height'] = self.height.to_dict()
        # override the default output from pydantic by calling `to_dict()` of width
        if self.width:
            _dict['width'] = self.width.to_dict()
        # override the default output from pydantic by calling `to_dict()` of seed
        if self.seed:
            _dict['seed'] = self.seed.to_dict()
        # override the default output from pydantic by calling `to_dict()` of controlnet_image
        if self.controlnet_image:
            _dict['controlnet_image'] = self.controlnet_image.to_dict()
        # override the default output from pydantic by calling `to_dict()` of init_image
        if self.init_image:
            _dict['init_image'] = self.init_image.to_dict()
        # override the default output from pydantic by calling `to_dict()` of mask_image
        if self.mask_image:
            _dict['mask_image'] = self.mask_image.to_dict()
        # override the default output from pydantic by calling `to_dict()` of style_preset
        if self.style_preset:
            _dict['style_preset'] = self.style_preset.to_dict()
        # puts key-value pairs in additional_properties in the top level
        if self.additional_properties is not None:
            for _key, _value in self.additional_properties.items():
                _dict[_key] = _value

        # set to None if prompt (nullable) is None
        # and __fields_set__ contains the field
        if self.prompt is None and "prompt" in self.__fields_set__:
            _dict['prompt'] = None

        # set to None if cfg_scale (nullable) is None
        # and __fields_set__ contains the field
        if self.cfg_scale is None and "cfg_scale" in self.__fields_set__:
            _dict['cfg_scale'] = None

        # set to None if steps (nullable) is None
        # and __fields_set__ contains the field
        if self.steps is None and "steps" in self.__fields_set__:
            _dict['steps'] = None

        # set to None if num_images (nullable) is None
        # and __fields_set__ contains the field
        if self.num_images is None and "num_images" in self.__fields_set__:
            _dict['num_images'] = None

        # set to None if strength (nullable) is None
        # and __fields_set__ contains the field
        if self.strength is None and "strength" in self.__fields_set__:
            _dict['strength'] = None

        # set to None if use_refiner (nullable) is None
        # and __fields_set__ contains the field
        if self.use_refiner is None and "use_refiner" in self.__fields_set__:
            _dict['use_refiner'] = None

        # set to None if high_noise_frac (nullable) is None
        # and __fields_set__ contains the field
        if self.high_noise_frac is None and "high_noise_frac" in self.__fields_set__:
            _dict['high_noise_frac'] = None

        # set to None if controlnet_conditioning_scale (nullable) is None
        # and __fields_set__ contains the field
        if self.controlnet_conditioning_scale is None and "controlnet_conditioning_scale" in self.__fields_set__:
            _dict['controlnet_conditioning_scale'] = None

        # set to None if enable_safety (nullable) is None
        # and __fields_set__ contains the field
        if self.enable_safety is None and "enable_safety" in self.__fields_set__:
            _dict['enable_safety'] = None

        # set to None if force_asset_download (nullable) is None
        # and __fields_set__ contains the field
        if self.force_asset_download is None and "force_asset_download" in self.__fields_set__:
            _dict['force_asset_download'] = None

        # set to None if force_asset_gpu_copy (nullable) is None
        # and __fields_set__ contains the field
        if self.force_asset_gpu_copy is None and "force_asset_gpu_copy" in self.__fields_set__:
            _dict['force_asset_gpu_copy'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ImageGenerationRequest:
        """Create an instance of ImageGenerationRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ImageGenerationRequest.parse_obj(obj)

        _obj = ImageGenerationRequest.parse_obj({
            "prompt": obj.get("prompt"),
            "prompt_2": SecondInputPrompt.from_dict(obj.get("prompt_2")) if obj.get("prompt_2") is not None else None,
            "negative_prompt": NegativePrompt.from_dict(obj.get("negative_prompt")) if obj.get("negative_prompt") is not None else None,
            "negative_prompt_2": SecondNegativePrompt.from_dict(obj.get("negative_prompt_2")) if obj.get("negative_prompt_2") is not None else None,
            "checkpoint": Checkpoint.from_dict(obj.get("checkpoint")) if obj.get("checkpoint") is not None else None,
            "controlnet": ControlNet.from_dict(obj.get("controlnet")) if obj.get("controlnet") is not None else None,
            "vae": VAE.from_dict(obj.get("vae")) if obj.get("vae") is not None else None,
            "textual_inversions": DictionaryOfTextualInversions.from_dict(obj.get("textual_inversions")) if obj.get("textual_inversions") is not None else None,
            "loras": DictionaryOfLoRAs.from_dict(obj.get("loras")) if obj.get("loras") is not None else None,
            "sampler": obj.get("sampler"),
            "height": OutputImageHeight.from_dict(obj.get("height")) if obj.get("height") is not None else None,
            "width": OutputImageWidth.from_dict(obj.get("width")) if obj.get("width") is not None else None,
            "cfg_scale": obj.get("cfg_scale"),
            "steps": obj.get("steps"),
            "num_images": obj.get("num_images"),
            "seed": FixedRandomSeed.from_dict(obj.get("seed")) if obj.get("seed") is not None else None,
            "controlnet_image": ControlNetImage.from_dict(obj.get("controlnet_image")) if obj.get("controlnet_image") is not None else None,
            "init_image": InitialImage.from_dict(obj.get("init_image")) if obj.get("init_image") is not None else None,
            "mask_image": MaskImage.from_dict(obj.get("mask_image")) if obj.get("mask_image") is not None else None,
            "strength": obj.get("strength"),
            "style_preset": PreDefinedStyles.from_dict(obj.get("style_preset")) if obj.get("style_preset") is not None else None,
            "use_refiner": obj.get("use_refiner"),
            "high_noise_frac": obj.get("high_noise_frac"),
            "controlnet_conditioning_scale": obj.get("controlnet_conditioning_scale"),
            "enable_safety": obj.get("enable_safety"),
            "image_encoding": obj.get("image_encoding"),
            "force_asset_download": obj.get("force_asset_download"),
            "force_asset_gpu_copy": obj.get("force_asset_gpu_copy")
        })
        # store additional fields in additional_properties
        for _key in obj.keys():
            if _key not in cls.__properties:
                _obj.additional_properties[_key] = obj.get(_key)

        return _obj





"""Example OctoAI service scaffold: Hello World."""
import pathlib

from octoai.service import Service, volume_path

ASSET_FILENAME: str = "asset.txt"


class HelloService(Service):
    """An OctoAI service extends octoai.service.Service."""

    def store_assets(self) -> None:
        """Download model assets."""
        root_path = pathlib.Path(volume_path())
        with open(root_path / ASSET_FILENAME, "w") as f:
            f.write("using volume")

    def setup(self):
        """Perform intialization."""
        print("Setting up.")
        root_path = pathlib.Path(volume_path())
        with open(root_path / ASSET_FILENAME, "r") as f:
            self.asset = f.read()

    def infer(self, prompt: str) -> str:
        """Perform inference."""
        return prompt + f" {self.asset}!"
